@@ -1,10 +1,29 @@
-from enum import Enum
-from typing import Any, Sequence
+from __future__ import annotations
 
+from collections.abc import Sequence
+from enum import Enum
+from typing import Any
+from warnings import warn
+
+from typing_extensions import deprecated
+
+# EmptyChannelError is re-exported from langgraph.channels.base
 from langgraph.checkpoint.base import EmptyChannelError  # noqa: F401
 from langgraph.types import Command, Interrupt
+from langgraph.warnings import LangGraphDeprecatedSinceV10
 
-# EmptyChannelError re-exported for backwards compatibility
+__all__ = (
+    "EmptyChannelError",
+    "ErrorCode",
+    "GraphRecursionError",
+    "InvalidUpdateError",
+    "GraphBubbleUp",
+    "GraphInterrupt",
+    "NodeInterrupt",
+    "ParentCommand",
+    "EmptyInputError",
+    "TaskNotFound",
+)
 
 
 class ErrorCode(Enum):
@@ -70,18 +89,26 @@ class GraphInterrupt(GraphBubbleUp):
         super().__init__(interrupts)
 
 
+@deprecated(
+    "NodeInterrupt is deprecated. Please use `langgraph.types.interrupt` instead.",
+    category=None,
+)
 class NodeInterrupt(GraphInterrupt):
-    """Raised by a node to interrupt execution."""
+    """Raised by a node to interrupt execution.
 
-    def __init__(self, value: Any) -> None:
-        super().__init__([Interrupt(value=value)])
+    Deprecated in V1.0.0 in favor of [`interrupt`][langgraph.types.interrupt].
+    """
 
-
-class GraphDelegate(GraphBubbleUp):
-    """Raised when a graph is delegated (for distributed mode)."""
-
-    def __init__(self, *args: dict[str, Any]) -> None:
-        super().__init__(*args)
+    def __init__(self, value: Any, id: str | None = None) -> None:
+        warn(
+            "NodeInterrupt is deprecated. Please use `langgraph.types.interrupt` instead.",
+            LangGraphDeprecatedSinceV10,
+            stacklevel=2,
+        )
+        if id is None:
+            super().__init__([Interrupt(value=value)])
+        else:
+            super().__init__([Interrupt(value=value, id=id)])
 
 
 class ParentCommand(GraphBubbleUp):
@@ -101,9 +128,3 @@ class TaskNotFound(Exception):
     """Raised when the executor is unable to find a task (for distributed mode)."""
 
     pass
-
-
-class CheckpointNotLatest(Exception):
-    """Raised when the checkpoint is not the latest version (for distributed mode)."""
-
-    pass
