@@ -1,13 +1,14 @@
+from __future__ import annotations
+
 import uuid
 import warnings
+from collections.abc import Sequence
 from functools import partial
 from typing import (
     Annotated,
     Any,
     Callable,
     Literal,
-    Optional,
-    Sequence,
     Union,
     cast,
 )
@@ -21,9 +22,17 @@ from langchain_core.messages import (
     convert_to_messages,
     message_chunk_to_message,
 )
-from typing_extensions import TypedDict
+from typing_extensions import TypedDict, deprecated
 
+from langgraph._internal._constants import CONF, CONFIG_KEY_SEND, NS_SEP
 from langgraph.graph.state import StateGraph
+from langgraph.warnings import LangGraphDeprecatedSinceV10
+
+__all__ = (
+    "add_messages",
+    "MessagesState",
+    "MessageGraph",
+)
 
 Messages = Union[list[MessageLikeRepresentation], MessageLikeRepresentation]
 
@@ -32,8 +41,8 @@ REMOVE_ALL_MESSAGES = "__remove_all__"
 
 def _add_messages_wrapper(func: Callable) -> Callable[[Messages, Messages], Messages]:
     def _add_messages(
-        left: Optional[Messages] = None, right: Optional[Messages] = None, **kwargs: Any
-    ) -> Union[Messages, Callable[[Messages, Messages], Messages]]:
+        left: Messages | None = None, right: Messages | None = None, **kwargs: Any
+    ) -> Messages | Callable[[Messages, Messages], Messages]:
         if left is not None and right is not None:
             return func(left, right, **kwargs)
         elif left is not None or right is not None:
@@ -54,7 +63,7 @@ def add_messages(
     left: Messages,
     right: Messages,
     *,
-    format: Optional[Literal["langchain-openai"]] = None,
+    format: Literal["langchain-openai"] | None = None,
 ) -> Messages:
     """Merges two lists of messages, updating existing messages by ID.
 
@@ -71,94 +80,99 @@ def add_messages(
             format, meaning contents can be string, 'text' blocks, or 'image_url' blocks
             and tool responses are returned as their own ToolMessages.
 
-            **REQUIREMENT**: Must have ``langchain-core>=0.3.11`` installed to use this
-            feature.
+            !!! important "Requirement"
+
+                Must have ``langchain-core>=0.3.11`` installed to use this feature.
 
     Returns:
         A new list of messages with the messages from `right` merged into `left`.
         If a message in `right` has the same ID as a message in `left`, the
         message from `right` will replace the message from `left`.
 
-    Examples:
-        ```pycon
-        >>> from langchain_core.messages import AIMessage, HumanMessage
-        >>> msgs1 = [HumanMessage(content="Hello", id="1")]
-        >>> msgs2 = [AIMessage(content="Hi there!", id="2")]
-        >>> add_messages(msgs1, msgs2)
-        [HumanMessage(content='Hello', id='1'), AIMessage(content='Hi there!', id='2')]
-
-        >>> msgs1 = [HumanMessage(content="Hello", id="1")]
-        >>> msgs2 = [HumanMessage(content="Hello again", id="1")]
-        >>> add_messages(msgs1, msgs2)
-        [HumanMessage(content='Hello again', id='1')]
-
-        >>> from typing import Annotated
-        >>> from typing_extensions import TypedDict
-        >>> from langgraph.graph import StateGraph
-        >>>
-        >>> class State(TypedDict):
-        ...     messages: Annotated[list, add_messages]
-        ...
-        >>> builder = StateGraph(State)
-        >>> builder.add_node("chatbot", lambda state: {"messages": [("assistant", "Hello")]})
-        >>> builder.set_entry_point("chatbot")
-        >>> builder.set_finish_point("chatbot")
-        >>> graph = builder.compile()
-        >>> graph.invoke({})
-        {'messages': [AIMessage(content='Hello', id=...)]}
-
-        >>> from typing import Annotated
-        >>> from typing_extensions import TypedDict
-        >>> from langgraph.graph import StateGraph, add_messages
-        >>>
-        >>> class State(TypedDict):
-        ...     messages: Annotated[list, add_messages(format='langchain-openai')]
-        ...
-        >>> def chatbot_node(state: State) -> list:
-        ...     return {"messages": [
-        ...         {
-        ...             "role": "user",
-        ...             "content": [
-        ...                 {
-        ...                     "type": "text",
-        ...                     "text": "Here's an image:",
-        ...                     "cache_control": {"type": "ephemeral"},
-        ...                 },
-        ...                 {
-        ...                     "type": "image",
-        ...                     "source": {
-        ...                         "type": "base64",
-        ...                         "media_type": "image/jpeg",
-        ...                         "data": "1234",
-        ...                     },
-        ...                 },
-        ...             ]
-        ...         },
-        ...     ]}
-        >>> builder = StateGraph(State)
-        >>> builder.add_node("chatbot", chatbot_node)
-        >>> builder.set_entry_point("chatbot")
-        >>> builder.set_finish_point("chatbot")
-        >>> graph = builder.compile()
-        >>> graph.invoke({"messages": []})
-        {
-            'messages': [
-                HumanMessage(
-                    content=[
-                        {"type": "text", "text": "Here's an image:"},
-                        {
-                            "type": "image_url",
-                            "image_url": {"url": "data:image/jpeg;base64,1234"},
-                        },
-                    ],
-                ),
-            ]
-        }
+    Example:
+        ```python title="Basic usage"
+        from langchain_core.messages import AIMessage, HumanMessage
+        msgs1 = [HumanMessage(content="Hello", id="1")]
+        msgs2 = [AIMessage(content="Hi there!", id="2")]
+        add_messages(msgs1, msgs2)
+        # [HumanMessage(content='Hello', id='1'), AIMessage(content='Hi there!', id='2')]
         ```
 
-    ..versionchanged:: 0.2.61
+        ```python title="Overwrite existing message"
+        msgs1 = [HumanMessage(content="Hello", id="1")]
+        msgs2 = [HumanMessage(content="Hello again", id="1")]
+        add_messages(msgs1, msgs2)
+        # [HumanMessage(content='Hello again', id='1')]
+        ```
+
+        ```python title="Use in a StateGraph"
+        from typing import Annotated
+        from typing_extensions import TypedDict
+        from langgraph.graph import StateGraph
+
+        class State(TypedDict):
+            messages: Annotated[list, add_messages]
+
+        builder = StateGraph(State)
+        builder.add_node("chatbot", lambda state: {"messages": [("assistant", "Hello")]})
+        builder.set_entry_point("chatbot")
+        builder.set_finish_point("chatbot")
+        graph = builder.compile()
+        graph.invoke({})
+        # {'messages': [AIMessage(content='Hello', id=...)]}
+        ```
+
+        ```python title="Use OpenAI message format"
+        from typing import Annotated
+        from typing_extensions import TypedDict
+        from langgraph.graph import StateGraph, add_messages
+
+        class State(TypedDict):
+            messages: Annotated[list, add_messages(format='langchain-openai')]
+
+        def chatbot_node(state: State) -> list:
+            return {"messages": [
+                {
+                    "role": "user",
+                    "content": [
+                        {
+                            "type": "text",
+                            "text": "Here's an image:",
+                            "cache_control": {"type": "ephemeral"},
+                        },
+                        {
+                            "type": "image",
+                            "source": {
+                                "type": "base64",
+                                "media_type": "image/jpeg",
+                                "data": "1234",
+                            },
+                        },
+                    ]
+                },
+            ]}
+
+        builder = StateGraph(State)
+        builder.add_node("chatbot", chatbot_node)
+        builder.set_entry_point("chatbot")
+        builder.set_finish_point("chatbot")
+        graph = builder.compile()
+        graph.invoke({"messages": []})
+        # {
+        #     'messages': [
+        #         HumanMessage(
+        #             content=[
+        #                 {"type": "text", "text": "Here's an image:"},
+        #                 {
+        #                     "type": "image_url",
+        #                     "image_url": {"url": "data:image/jpeg;base64,1234"},
+        #                 },
+        #             ],
+        #         ),
+        #     ]
+        # }
+        ```
 
-        Support for 'format="langchain-openai"' flag added.
     """
     remove_all_idx = None
     # coerce to list
@@ -220,9 +234,16 @@ def add_messages(
     return merged
 
 
+@deprecated(
+    "MessageGraph is deprecated in LangGraph v1.0.0, to be removed in v2.0.0. Please use StateGraph with a `messages` key instead.",
+    category=None,
+)
 class MessageGraph(StateGraph):
     """A StateGraph where every node receives a list of messages as input and returns one or more messages as output.
 
+    !!! warning "Deprecation"
+        MessageGraph is deprecated in LangGraph v1.0.0, to be removed in v2.0.0. Please use StateGraph with a `messages` key instead.
+
     MessageGraph is a subclass of StateGraph whose entire state is a single, append-only* list of messages.
     Each node in a MessageGraph takes a list of messages as input and returns zero or more
     messages as output. The `add_messages` function is used to merge the output messages from each node
@@ -268,6 +289,11 @@ class MessageGraph(StateGraph):
     """
 
     def __init__(self) -> None:
+        warnings.warn(
+            "MessageGraph is deprecated in LangGraph v1.0.0, to be removed in v2.0.0. Please use StateGraph with a `messages` key instead.",
+            category=LangGraphDeprecatedSinceV10,
+            stacklevel=2,
+        )
         super().__init__(Annotated[list[AnyMessage], add_messages])  # type: ignore[arg-type]
 
 
@@ -289,3 +315,51 @@ def _format_messages(messages: Sequence[BaseMessage]) -> list[BaseMessage]:
         return list(messages)
     else:
         return convert_to_messages(convert_to_openai_messages(messages))
+
+
+def push_message(
+    message: MessageLikeRepresentation | BaseMessageChunk,
+    *,
+    state_key: str | None = "messages",
+) -> AnyMessage:
+    """Write a message manually to the `messages` / `messages-tuple` stream mode.
+
+    Will automatically write to the channel specified in the `state_key` unless `state_key` is `None`.
+    """
+
+    from langchain_core.callbacks.base import (
+        BaseCallbackHandler,
+        BaseCallbackManager,
+    )
+
+    from langgraph.config import get_config
+    from langgraph.pregel._messages import StreamMessagesHandler
+
+    config = get_config()
+    message = next(x for x in convert_to_messages([message]))
+
+    if message.id is None:
+        raise ValueError("Message ID is required")
+
+    if isinstance(config["callbacks"], BaseCallbackManager):
+        manager = config["callbacks"]
+        handlers = manager.handlers
+    elif isinstance(config["callbacks"], list) and all(
+        isinstance(x, BaseCallbackHandler) for x in config["callbacks"]
+    ):
+        handlers = config["callbacks"]
+
+    if stream_handler := next(
+        (x for x in handlers if isinstance(x, StreamMessagesHandler)), None
+    ):
+        metadata = config["metadata"]
+        message_meta = (
+            tuple(cast(str, metadata["langgraph_checkpoint_ns"]).split(NS_SEP)),
+            metadata,
+        )
+        stream_handler._emit(message_meta, message, dedupe=False)
+
+    if state_key:
+        config[CONF][CONFIG_KEY_SEND]([(state_key, message)])
+
+    return message
