@@ -1,32 +1,36 @@
-import dataclasses
+from __future__ import annotations
+
 import sys
 from collections import deque
+from collections.abc import Hashable, Sequence
+from dataclasses import asdict, dataclass
 from typing import (
     TYPE_CHECKING,
     Any,
     Callable,
     ClassVar,
     Generic,
-    Hashable,
     Literal,
     NamedTuple,
-    Optional,
-    Sequence,
-    Type,
     TypeVar,
     Union,
-    cast,
-    get_type_hints,
+    final,
 )
+from warnings import warn
 
 from langchain_core.runnables import Runnable, RunnableConfig
-from typing_extensions import Self
+from typing_extensions import Unpack, deprecated
+from xxhash import xxh3_128_hexdigest
 
+from langgraph._internal._cache import default_cache_key
+from langgraph._internal._fields import get_cached_annotated_keys, get_update_as_tuples
+from langgraph._internal._retry import default_retry_on
+from langgraph._internal._typing import MISSING, DeprecatedKwargs
 from langgraph.checkpoint.base import BaseCheckpointSaver, CheckpointMetadata
+from langgraph.warnings import LangGraphDeprecatedSinceV10
 
 if TYPE_CHECKING:
     from langgraph.pregel.protocol import PregelProtocol
-    from langgraph.store.base import BaseStore
 
 
 try:
@@ -37,6 +41,30 @@ except ImportError:
         pass
 
 
+__all__ = (
+    "All",
+    "Checkpointer",
+    "StreamMode",
+    "StreamWriter",
+    "RetryPolicy",
+    "CachePolicy",
+    "Interrupt",
+    "StateUpdate",
+    "PregelTask",
+    "PregelExecutableTask",
+    "StateSnapshot",
+    "Send",
+    "Command",
+    "Durability",
+    "interrupt",
+)
+
+Durability = Literal["sync", "async", "exit"]
+"""Durability mode for the graph execution.
+- `"sync"`: Changes are persisted synchronously before the next step starts.
+- `"async"`: Changes are persisted asynchronously while the next step executes.
+- `"exit"`: Changes are persisted only when the graph exits."""
+
 All = Literal["*"]
 """Special value to indicate that graph should interrupt on all nodes."""
 
@@ -46,16 +74,20 @@ Checkpointer = Union[None, bool, BaseCheckpointSaver]
 - False disables checkpointing, even if the parent graph has a checkpointer.
 - None inherits checkpointer from the parent graph."""
 
-StreamMode = Literal["values", "updates", "debug", "messages", "custom"]
+StreamMode = Literal[
+    "values", "updates", "checkpoints", "tasks", "debug", "messages", "custom"
+]
 """How the stream method should emit outputs.
 
-- `"values"`: Emit all values in the state after each step.
+- `"values"`: Emit all values in the state after each step, including interrupts.
     When used with functional API, values are emitted once at the end of the workflow.
 - `"updates"`: Emit only the node or task names and updates returned by the nodes or tasks after each step.
     If multiple updates are made in the same step (e.g. multiple nodes are run) then those updates are emitted separately.
 - `"custom"`: Emit custom data using from inside nodes or tasks using `StreamWriter`.
 - `"messages"`: Emit LLM messages token-by-token together with metadata for any LLM invocations inside nodes or tasks.
-- `"debug"`: Emit debug events with as much information as possible for each step.
+- `"checkpoints"`: Emit an event when a checkpoint is created, in the same format as returned by get_state().
+- `"tasks"`: Emit events when tasks start and finish, including their results and errors.
+- `"debug"`: Emit "checkpoints" and "tasks" events, for debugging purposes.
 """
 
 StreamWriter = Callable[[Any], None]
@@ -64,42 +96,13 @@ Always injected into nodes if requested as a keyword argument, but it's a no-op
 when not using stream_mode="custom"."""
 
 if sys.version_info >= (3, 10):
+    _DC_SLOTS = {"slots": True}
     _DC_KWARGS = {"kw_only": True, "slots": True, "frozen": True}
 else:
+    _DC_SLOTS = {}
     _DC_KWARGS = {"frozen": True}
 
 
-def default_retry_on(exc: Exception) -> bool:
-    import httpx
-    import requests
-
-    if isinstance(exc, ConnectionError):
-        return True
-    if isinstance(
-        exc,
-        (
-            ValueError,
-            TypeError,
-            ArithmeticError,
-            ImportError,
-            LookupError,
-            NameError,
-            SyntaxError,
-            RuntimeError,
-            ReferenceError,
-            StopIteration,
-            StopAsyncIteration,
-            OSError,
-        ),
-    ):
-        return False
-    if isinstance(exc, httpx.HTTPStatusError):
-        return 500 <= exc.response.status_code < 600
-    if isinstance(exc, requests.HTTPError):
-        return 500 <= exc.response.status_code < 600 if exc.response else True
-    return True
-
-
 class RetryPolicy(NamedTuple):
     """Configuration for retrying nodes.
 
@@ -116,46 +119,104 @@ class RetryPolicy(NamedTuple):
     """Maximum number of attempts to make before giving up, including the first."""
     jitter: bool = True
     """Whether to add random jitter to the interval between retries."""
-    retry_on: Union[
-        Type[Exception], Sequence[Type[Exception]], Callable[[Exception], bool]
-    ] = default_retry_on
+    retry_on: (
+        type[Exception] | Sequence[type[Exception]] | Callable[[Exception], bool]
+    ) = default_retry_on
     """List of exception classes that should trigger a retry, or a callable that returns True for exceptions that should trigger a retry."""
 
 
-class CachePolicy(NamedTuple):
-    """Configuration for caching nodes.
-
-    !!! version-added "Added in version 0.2.24."
-    """
-
-    pass
+KeyFuncT = TypeVar("KeyFuncT", bound=Callable[..., Union[str, bytes]])
 
 
-@dataclasses.dataclass(**_DC_KWARGS)
+@dataclass(**_DC_KWARGS)
+class CachePolicy(Generic[KeyFuncT]):
+    """Configuration for caching nodes."""
+
+    key_func: KeyFuncT = default_cache_key  # type: ignore[assignment]
+    """Function to generate a cache key from the node's input.
+    Defaults to hashing the input with pickle."""
+
+    ttl: int | None = None
+    """Time to live for the cache entry in seconds. If None, the entry never expires."""
+
+
+_DEFAULT_INTERRUPT_ID = "placeholder-id"
+
+
+@final
+@dataclass(init=False, **_DC_SLOTS)
 class Interrupt:
-    """
+    """Information about an interrupt that occurred in a node.
+
     !!! version-added "Added in version 0.2.24."
+
+    !!! version-changed "Changed in version v0.4.0"
+        * `interrupt_id` was introduced as a property
+
+    !!! version-changed "Changed in version v0.6.0"
+
+        The following attributes have been removed:
+
+        * `ns`
+        * `when`
+        * `resumable`
+        * `interrupt_id`, deprecated in favor of `id`
     """
 
     value: Any
-    resumable: bool = False
-    ns: Optional[Sequence[str]] = None
-    when: Literal["during"] = dataclasses.field(default="during", repr=False)
+    """The value associated with the interrupt."""
+
+    id: str
+    """The ID of the interrupt. Can be used to resume the interrupt directly."""
+
+    def __init__(
+        self,
+        value: Any,
+        id: str = _DEFAULT_INTERRUPT_ID,
+        **deprecated_kwargs: Unpack[DeprecatedKwargs],
+    ) -> None:
+        self.value = value
+
+        if (
+            (ns := deprecated_kwargs.get("ns", MISSING)) is not MISSING
+            and (id == _DEFAULT_INTERRUPT_ID)
+            and (isinstance(ns, Sequence))
+        ):
+            self.id = xxh3_128_hexdigest("|".join(ns).encode())
+        else:
+            self.id = id
+
+    @classmethod
+    def from_ns(cls, value: Any, ns: str) -> Interrupt:
+        return cls(value=value, id=xxh3_128_hexdigest(ns.encode()))
+
+    @property
+    @deprecated("`interrupt_id` is deprecated. Use `id` instead.", category=None)
+    def interrupt_id(self) -> str:
+        warn(
+            "`interrupt_id` is deprecated. Use `id` instead.",
+            LangGraphDeprecatedSinceV10,
+            stacklevel=2,
+        )
+        return self.id
 
 
 class StateUpdate(NamedTuple):
-    values: Optional[dict[str, Any]]
-    as_node: Optional[str] = None
+    values: dict[str, Any] | None
+    as_node: str | None = None
+    task_id: str | None = None
 
 
 class PregelTask(NamedTuple):
+    """A Pregel task."""
+
     id: str
     name: str
-    path: tuple[Union[str, int, tuple], ...]
-    error: Optional[Exception] = None
+    path: tuple[str | int | tuple, ...]
+    error: Exception | None = None
     interrupts: tuple[Interrupt, ...] = ()
-    state: Union[None, RunnableConfig, "StateSnapshot"] = None
-    result: Optional[Any] = None
+    state: None | RunnableConfig | StateSnapshot = None
+    result: Any | None = None
 
 
 if sys.version_info > (3, 11):
@@ -164,7 +225,18 @@ else:
     _T_DC_KWARGS = {"frozen": True}
 
 
-@dataclasses.dataclass(**_T_DC_KWARGS)
+class CacheKey(NamedTuple):
+    """Cache key for a task."""
+
+    ns: tuple[str, ...]
+    """Namespace for the cache entry."""
+    key: str
+    """Key for the cache entry."""
+    ttl: int | None
+    """Time to live for the cache entry in seconds."""
+
+
+@dataclass(**_T_DC_KWARGS)
 class PregelExecutableTask:
     name: str
     input: Any
@@ -172,32 +244,33 @@ class PregelExecutableTask:
     writes: deque[tuple[str, Any]]
     config: RunnableConfig
     triggers: Sequence[str]
-    retry_policy: Optional[RetryPolicy]
-    cache_policy: Optional[CachePolicy]
+    retry_policy: Sequence[RetryPolicy]
+    cache_key: CacheKey | None
     id: str
-    path: tuple[Union[str, int, tuple], ...]
-    scheduled: bool = False
+    path: tuple[str | int | tuple, ...]
     writers: Sequence[Runnable] = ()
-    subgraphs: Sequence["PregelProtocol"] = ()
+    subgraphs: Sequence[PregelProtocol] = ()
 
 
 class StateSnapshot(NamedTuple):
     """Snapshot of the state of the graph at the beginning of a step."""
 
-    values: Union[dict[str, Any], Any]
-    """Current values of channels"""
+    values: dict[str, Any] | Any
+    """Current values of channels."""
     next: tuple[str, ...]
     """The name of the node to execute in each task for this step."""
     config: RunnableConfig
-    """Config used to fetch this snapshot"""
-    metadata: Optional[CheckpointMetadata]
-    """Metadata associated with this snapshot"""
-    created_at: Optional[str]
-    """Timestamp of snapshot creation"""
-    parent_config: Optional[RunnableConfig]
-    """Config used to fetch the parent snapshot, if any"""
+    """Config used to fetch this snapshot."""
+    metadata: CheckpointMetadata | None
+    """Metadata associated with this snapshot."""
+    created_at: str | None
+    """Timestamp of snapshot creation."""
+    parent_config: RunnableConfig | None
+    """Config used to fetch the parent snapshot, if any."""
     tasks: tuple[PregelTask, ...]
     """Tasks to execute in this step. If already attempted, may contain an error."""
+    interrupts: tuple[Interrupt, ...]
+    """Interrupts that occurred in this step that are pending resolution."""
 
 
 class Send:
@@ -251,8 +324,8 @@ class Send:
         Initialize a new instance of the Send class.
 
         Args:
-            node (str): The name of the target node to send the message to.
-            arg (Any): The state or message to send to the target node.
+            node: The name of the target node to send the message to.
+            arg: The state or message to send to the target node.
         """
         self.node = node
         self.arg = arg
@@ -274,7 +347,7 @@ class Send:
 N = TypeVar("N", bound=Hashable)
 
 
-@dataclasses.dataclass(**_DC_KWARGS)
+@dataclass(**_DC_KWARGS)
 class Command(Generic[N], ToolOutputMixin):
     """One or more commands to update the graph's state and send messages to nodes.
 
@@ -287,6 +360,10 @@ class Command(Generic[N], ToolOutputMixin):
             - Command.PARENT: closest parent graph
         update: update to apply to the graph's state.
         resume: value to resume execution with. To be used together with [`interrupt()`][langgraph.types.interrupt].
+            Can be one of the following:
+
+            - mapping of interrupt ids to resume values
+            - a single value with which to resume the next interrupt
         goto: can be one of the following:
 
             - name of the node to navigate to next (any node that belongs to the specified `graph`)
@@ -295,17 +372,15 @@ class Command(Generic[N], ToolOutputMixin):
             - sequence of `Send` objects
     """
 
-    graph: Optional[str] = None
-    update: Optional[Any] = None
-    resume: Optional[Union[Any, dict[str, Any]]] = None
-    goto: Union[Send, Sequence[Union[Send, str]], str] = ()
+    graph: str | None = None
+    update: Any | None = None
+    resume: dict[str, Any] | Any | None = None
+    goto: Send | Sequence[Send | N] | N = ()
 
     def __repr__(self) -> str:
         # get all non-None values
         contents = ", ".join(
-            f"{key}={value!r}"
-            for key, value in dataclasses.asdict(self).items()
-            if value
+            f"{key}={value!r}" for key, value in asdict(self).items() if value
         )
         return f"Command({contents})"
 
@@ -317,8 +392,8 @@ class Command(Generic[N], ToolOutputMixin):
             for t in self.update
         ):
             return self.update
-        elif hints := get_type_hints(type(self.update)):
-            return [(k, getattr(self.update, k)) for k in hints]
+        elif keys := get_cached_annotated_keys(type(self.update)):
+            return get_update_as_tuples(self.update, keys)
         elif self.update is not None:
             return [("__root__", self.update)]
         else:
@@ -327,60 +402,6 @@ class Command(Generic[N], ToolOutputMixin):
     PARENT: ClassVar[Literal["__parent__"]] = "__parent__"
 
 
-StreamChunk = tuple[tuple[str, ...], str, Any]
-
-
-class StreamProtocol:
-    __slots__ = ("modes", "__call__")
-
-    modes: set[StreamMode]
-
-    __call__: Callable[[Self, StreamChunk], None]
-
-    def __init__(
-        self,
-        __call__: Callable[[StreamChunk], None],
-        modes: set[StreamMode],
-    ) -> None:
-        self.__call__ = cast(Callable[[Self, StreamChunk], None], __call__)
-        self.modes = modes
-
-
-class LoopProtocol:
-    config: RunnableConfig
-    store: Optional["BaseStore"]
-    stream: Optional[StreamProtocol]
-    step: int
-    stop: int
-
-    def __init__(
-        self,
-        *,
-        step: int,
-        stop: int,
-        config: RunnableConfig,
-        store: Optional["BaseStore"] = None,
-        stream: Optional[StreamProtocol] = None,
-    ) -> None:
-        self.stream = stream
-        self.config = config
-        self.store = store
-        self.step = step
-        self.stop = stop
-
-
-@dataclasses.dataclass(**_DC_KWARGS)
-class PregelScratchpad:
-    # call
-    call_counter: Callable[[], int]
-    # interrupt
-    interrupt_counter: Callable[[], int]
-    get_null_resume: Callable[[bool], Any]
-    resume: list[Any]
-    # subgraph
-    subgraph_counter: Callable[[], int]
-
-
 def interrupt(value: Any) -> Any:
     """Interrupt the graph with a resumable exception from within a node.
 
@@ -409,10 +430,10 @@ def interrupt(value: Any) -> Any:
         from typing import Optional
         from typing_extensions import TypedDict
 
-        from langgraph.checkpoint.memory import MemorySaver
+        from langgraph.checkpoint.memory import InMemorySaver
         from langgraph.constants import START
         from langgraph.graph import StateGraph
-        from langgraph.types import interrupt
+        from langgraph.types import interrupt, Command
 
 
         class State(TypedDict):
@@ -438,7 +459,7 @@ def interrupt(value: Any) -> Any:
         builder.add_edge(START, \"node\")
 
         # A checkpointer must be enabled for interrupts to work!
-        checkpointer = MemorySaver()
+        checkpointer = InMemorySaver()
         graph = builder.compile(checkpointer=checkpointer)
 
         config = {
@@ -449,22 +470,16 @@ def interrupt(value: Any) -> Any:
 
         for chunk in graph.stream({\"foo\": \"abc\"}, config):
             print(chunk)
-        ```
 
-        ```pycon
-        {'__interrupt__': (Interrupt(value='what is your age?', resumable=True, ns=['node:62e598fa-8653-9d6d-2046-a70203020e37'], when='during'),)}
-        ```
+        # > {'__interrupt__': (Interrupt(value='what is your age?', id='45fda8478b2ef754419799e10992af06'),)}
 
-        ```python
         command = Command(resume=\"some input from a human!!!\")
 
         for chunk in graph.stream(Command(resume=\"some input from a human!!!\"), config):
             print(chunk)
-        ```
 
-        ```pycon
-        Received an input from the interrupt: some input from a human!!!
-        {'node': {'human_value': 'some input from a human!!!'}}
+        # > Received an input from the interrupt: some input from a human!!!
+        # > {'node': {'human_value': 'some input from a human!!!'}}
         ```
 
     Args:
@@ -476,19 +491,18 @@ def interrupt(value: Any) -> Any:
     Raises:
         GraphInterrupt: On the first invocation within the node, halts execution and surfaces the provided value to the client.
     """
-    from langgraph.constants import (
+    from langgraph._internal._constants import (
         CONFIG_KEY_CHECKPOINT_NS,
         CONFIG_KEY_SCRATCHPAD,
         CONFIG_KEY_SEND,
-        NS_SEP,
         RESUME,
     )
+    from langgraph.config import get_config
     from langgraph.errors import GraphInterrupt
-    from langgraph.utils.config import get_config
 
     conf = get_config()["configurable"]
     # track interrupt index
-    scratchpad: PregelScratchpad = conf[CONFIG_KEY_SCRATCHPAD]
+    scratchpad = conf[CONFIG_KEY_SCRATCHPAD]
     idx = scratchpad.interrupt_counter()
     # find previous resume values
     if scratchpad.resume:
@@ -504,10 +518,9 @@ def interrupt(value: Any) -> Any:
     # no resume value found
     raise GraphInterrupt(
         (
-            Interrupt(
+            Interrupt.from_ns(
                 value=value,
-                resumable=True,
-                ns=cast(str, conf[CONFIG_KEY_CHECKPOINT_NS]).split(NS_SEP),
+                ns=conf[CONFIG_KEY_CHECKPOINT_NS],
             ),
         )
     )
